"""Turn your Excel calculators into python functions."""

__version__ = "0.0.1"

from functools import wraps
import xlwings as xw
import collections.abc


class _Calculation:
    """Object used to work with the Excel ranges and get the calculations."""

    _input_rng_list: list[xw.main.Range] = None
    _output_rng: xw.main.Range = None

    @property
    def input_rng_list(self):
        return self._input_rng_list

    @input_rng_list.setter
    def input_rng_list(self, value):
        self._input_rng_list = []
        for input_rng in value:
            if not isinstance(input_rng, xw.main.Range):
                raise TypeError(
                    f"Range objects are required input_rng_list, not {type(input_rng)}"
                )
            self._input_rng_list.append(input_rng)

    @property
    def output_rng(self):
        return self._output_rng

    @output_rng.setter
    def output_rng(self, value):
        if not isinstance(value, xw.main.Range):
            raise TypeError("Range object required for output_rng")
        self._output_rng = value

    def apply(self, *args, **kwargs):
        """Write the input args to the input ranges."""

        if kwargs:
            raise NotImplementedError("kwargs in calculation not supported")
        for arg, input_rng in zip(args, self.input_rng_list, strict=True):
            shape = input_rng.shape
            if len(shape) != 2:
                raise RuntimeError(f"unexpected type of range shape for {shape}")
            if not isinstance(arg, str) and isinstance(arg, collections.abc.Collection):
                if input_rng.shape == (1, 1):
                    # NOTE: xlwings already elegantly handles the case that arg is collection of shape (1,1)
                    raise TypeError(
                        f"not allowed to assign  type {type(arg).__name__} to singe cell {input_rng.address} because "
                        f"it contains multiple values"
                    )
                elif input_rng.shape[1] == 1:
                    # convert argument sequence to be compatible with a single column
                    input_rng.value = [[v] for v in arg]
                elif input_rng.shape[0] == 1:
                    # argument sequence is a single row
                    input_rng.value = [*arg]
                else:
                    # argument sequence needs to be 2d and not a 1d sequence of strings, nor a dict
                    # checking dict bc xlwings allows dicts; but see no harm in allowing dicts in other cases above
                    rows = []
                    for row in args:
                        if isinstance(row, str) or isinstance(row, dict):
                            raise TypeError(
                                f"2d range {input_rng.address} not compatible with 1d input {row}"
                            )
                        rows.append([*row])
                    input_rng.value = rows
            else:
                # single argument
                input_rng.value = arg

    def retrieve(self):
        """Get the output from the output range."""
        return self.output_rng.value

    def __call__(self, *args, **kwargs):
        self.apply(*args, **kwargs)
        return self.retrieve()


def create_calculation(input_rng_list, output_rng):
    """Decorator for turning Excel apps into functions.

    The functions use Excel in the background instead of computation being completed in python.
    """

    calc = _Calculation()
    calc.input_rng_list = input_rng_list
    calc.output_rng = output_rng

    def wrapper(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            return calc(*args, **kwargs)

        wrapped.calculation = calc
        return wrapped

    return wrapper


def adapt_function(input_adapter_list, output_adapter):
    """Decorator that allows modification of the inputs and outputs."""

    def wrapper(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            args = [
                input_adapter(arg)
                for input_adapter, arg in zip(input_adapter_list, args, strict=True)
            ]
            result = func(*args, **kwargs)
            return output_adapter(result)

        return wrapped

    return wrapper



from excalc_py import create_calculation, adapt_function
import pytest


@pytest.fixture
def decorated_func(input_rng1, input_rng2, output_rng, input_adapter, output_adapter):
    @adapt_function([input_adapter, input_adapter], output_adapter)
    @create_calculation([input_rng1, input_rng2], output_rng)
    def func(a, b):
        pass

    return func


@pytest.fixture
def inputs():
    return [1, 2]


@pytest.fixture
def expected_output():
    return {6}


def test_combined_decorators(decorated_func, inputs, expected_output):
    assert decorated_func(*inputs) == expected_output



import pytest
from excalc_py import adapt_function


@pytest.fixture
def inputs():
    return [1, 1]


@pytest.fixture
def expected_result():
    return {2}


@pytest.fixture
def my_func_calculation(input_adapter, output_adapter):
    @adapt_function([input_adapter, input_adapter], output_adapter)
    def my_func(a, b):
        return a, b

    return my_func


def test_adapt_for_units(my_func_calculation, inputs, expected_result):
    assert my_func_calculation(*inputs) == expected_result



import pytest
import faulthandler
import xlwings as xw
import collections.abc

TEST_WB = "test_book.xlsx"


@pytest.fixture(scope="session")
def xw_app():
    # setup
    faulthandler.disable()  # required for testing xlwings with pytest
    app = xw.App(visible=False)
    yield app
    # teardown
    app.quit()


@pytest.fixture
def wb(xw_app):
    return xw_app.books.open(TEST_WB)


@pytest.fixture
def input_rng1(wb):
    return wb.sheets[0].range("B1")


@pytest.fixture
def input_rng2(wb):
    return wb.sheets[0].range("B2")


@pytest.fixture
def output_rng(wb):
    return wb.sheets[0].range("B3")


@pytest.fixture
def input_adapter():
    def multiply_by_2(value):
        return value * 2

    return multiply_by_2


@pytest.fixture
def output_adapter():
    def make_set(value):
        if not isinstance(value, str) and isinstance(value, collections.abc.Iterable):
            return set(value)
        else:
            return {value}

    return make_set



from excalc_py import create_calculation, _Calculation
import pytest


@pytest.fixture
def inputs():
    return [1.1, 2.2]


@pytest.fixture
def expected_result():
    return 3.3


@pytest.fixture
def calculation():
    return _Calculation()


@pytest.fixture
def my_func_calculation(input_rng1, input_rng2, output_rng):
    @create_calculation([input_rng1, input_rng2], output_rng)
    def my_func(a, b):
        pass

    return my_func


def test_calculation_apply(calculation, input_rng1, input_rng2, inputs):
    calculation.input_rng_list = [input_rng1, input_rng2]
    calculation.apply(*inputs)
    assert (
        calculation.input_rng_list[0].sheet.range(input_rng1.address).value == inputs[0]
    )
    assert (
        calculation.input_rng_list[1].sheet.range(input_rng2.address).value == inputs[1]
    )


def test_create_calculation(my_func_calculation, inputs, expected_result):
    assert my_func_calculation(*inputs) == pytest.approx(expected_result)


def test_create_calculation_no_kwargs(my_func_calculation, inputs):
    with pytest.raises(TypeError):
        my_func_calculation(dict(zip("ab", inputs)))
